@@ -4,23 +4,19 @@ use std::{io, net::SocketAddr};
 
 use ntp_proto::NtpTimestamp;
 use tokio::io::{unix::AsyncFd, Interest};
-use tracing::{debug, instrument, trace, warn};
+use tracing::instrument;
 
 use crate::{
+    interface::InterfaceName,
     raw_socket::{
         control_message_space, receive_message, set_timestamping_options, ControlMessage,
         MessageQueue, TimestampMethod,
     },
-    EnableTimestamps, InterfaceName,
+    EnableTimestamps,
 };
 
-#[cfg(target_os = "linux")]
-use crate::raw_socket::err_queue_waiter::ErrQueueWaiter;
-
 pub struct UdpSocket {
     io: AsyncFd<std::net::UdpSocket>,
-    #[cfg(target_os = "linux")]
-    err_queue_waiter: ErrQueueWaiter,
     send_counter: u32,
     timestamping: EnableTimestamps,
 }
@@ -67,8 +63,8 @@ impl UdpSocket {
         timestamping: EnableTimestamps,
     ) -> io::Result<UdpSocket> {
         let socket = tokio::net::UdpSocket::bind(listen_addr).await?;
-        debug!(
-            local_addr = debug(socket.local_addr().unwrap()),
+        tracing::debug!(
+            local_addr = ?socket.local_addr().unwrap(),
             "client socket bound"
         );
 
@@ -80,9 +76,9 @@ impl UdpSocket {
         }
 
         socket.connect(peer_addr).await?;
-        debug!(
-            local_addr = debug(socket.local_addr().unwrap()),
-            peer_addr = debug(socket.peer_addr().unwrap()),
+        tracing::debug!(
+            local_addr = ?socket.local_addr().unwrap(),
+            peer_addr = ?socket.peer_addr().unwrap(),
             "client socket connected"
         );
 
@@ -91,8 +87,6 @@ impl UdpSocket {
         set_timestamping_options(&socket, method, timestamping)?;
 
         Ok(UdpSocket {
-            #[cfg(target_os = "linux")]
-            err_queue_waiter: ErrQueueWaiter::new(&socket)?,
             io: AsyncFd::new(socket)?,
             send_counter: 0,
             timestamping,
@@ -105,8 +99,8 @@ impl UdpSocket {
         interface: Option<InterfaceName>,
     ) -> io::Result<UdpSocket> {
         let socket = tokio::net::UdpSocket::bind(listen_addr).await?;
-        debug!(
-            local_addr = debug(socket.local_addr().unwrap()),
+        tracing::debug!(
+            local_addr = ?socket.local_addr().unwrap(),
             "server socket bound"
         );
 
@@ -131,8 +125,6 @@ impl UdpSocket {
         set_timestamping_options(&socket, DEFAULT_TIMESTAMP_METHOD, timestamping)?;
 
         Ok(UdpSocket {
-            #[cfg(target_os = "linux")]
-            err_queue_waiter: ErrQueueWaiter::new(&socket)?,
             io: AsyncFd::new(socket)?,
             send_counter: 0,
             timestamping,
@@ -145,7 +137,7 @@ impl UdpSocket {
         buf_size = buf.len(),
     ))]
     pub async fn send(&mut self, buf: &[u8]) -> io::Result<(usize, Option<NtpTimestamp>)> {
-        trace!(size = buf.len(), "sending bytes");
+        tracing::trace!(size = buf.len(), "sending bytes");
 
         let result = self
             .io
@@ -154,11 +146,11 @@ impl UdpSocket {
 
         let send_size = match result {
             Ok(size) => {
-                trace!(sent = size, "sent bytes");
+                tracing::trace!(sent = size, "sent bytes");
                 size
             }
             Err(e) => {
-                debug!(error = debug(&e), "error sending data");
+                tracing::debug!(error = debug(&e), "error sending data");
                 return Err(e);
             }
         };
@@ -178,7 +170,7 @@ impl UdpSocket {
                     .await
                 {
                     Err(_) => {
-                        warn!("Packet without timestamp");
+                        tracing::warn!("Packet without timestamp");
                         Ok((send_size, None))
                     }
                     Ok(send_timestamp) => Ok((send_size, Some(send_timestamp?))),
@@ -191,40 +183,26 @@ impl UdpSocket {
                 Ok((send_size, None))
             }
         } else {
-            trace!("send timestamping not supported");
+            tracing::trace!("send timestamping not supported");
             Ok((send_size, None))
         }
     }
 
     #[cfg(target_os = "linux")]
     async fn fetch_send_timestamp(&self, expected_counter: u32) -> io::Result<NtpTimestamp> {
-        trace!("waiting for timestamp socket to become readable to fetch a send timestamp");
+        let msg = "waiting for timestamp socket to become readable to fetch a send timestamp";
+        tracing::trace!(msg);
+
+        let try_read = |udp_socket: &std::net::UdpSocket| {
+            fetch_send_timestamp_help(udp_socket, expected_counter)
+        };
 
-        // Send timestamps are sent to the udp socket's error queue. Sadly, tokio does not
-        // currently support awaiting whether there is something in the error queue
-        // see https://github.com/tokio-rs/tokio/issues/4885.
-        //
-        // Therefore, we manually configure an extra file descriptor to listen for POLLPRI on
-        // the main udp socket. This `exceptional_condition` file descriptor becomes readable
-        // when there is something in the error queue.
         loop {
-            // Send timestamps are sent to the udp socket's error queue. Sadly, tokio does not
-            // currently support awaiting whether there is something in the error queue
-            // see https://github.com/tokio-rs/tokio/issues/4885.
-            self.err_queue_waiter.wait().await?;
-
-            match fetch_send_timestamp_help(self.io.get_ref(), expected_counter) {
-                Ok(Some(send_timestamp)) => {
-                    return Ok(send_timestamp);
-                }
-                Ok(None) => {
-                    continue;
-                }
-                Err(e) => {
-                    warn!(error = debug(&e), "Error fetching timestamp");
-                    return Err(e);
-                }
-            }
+            // the timestamp being available triggers the error interest
+            match self.io.async_io(Interest::ERROR, try_read).await? {
+                Some(timestamp) => return Ok(timestamp),
+                None => continue,
+            };
         }
     }
 
@@ -233,7 +211,7 @@ impl UdpSocket {
         buf_size = buf.len(),
     ))]
     pub async fn send_to(&self, buf: &[u8], addr: SocketAddr) -> io::Result<usize> {
-        trace!(size = buf.len(), ?addr, "sending bytes");
+        tracing::trace!(size = buf.len(), ?addr, "sending bytes");
 
         let result = self
             .io
@@ -241,8 +219,8 @@ impl UdpSocket {
             .await;
 
         match &result {
-            Ok(size) => trace!(sent = size, "sent bytes"),
-            Err(e) => debug!(error = debug(e), "error sending data"),
+            Ok(size) => tracing::trace!(sent = size, "sent bytes"),
+            Err(e) => tracing::debug!(error = debug(e), "error sending data"),
         }
 
         result
@@ -257,7 +235,7 @@ impl UdpSocket {
         &self,
         buf: &mut [u8],
     ) -> io::Result<(usize, SocketAddr, Option<NtpTimestamp>)> {
-        trace!("waiting for socket to become readable");
+        tracing::trace!("waiting for socket to become readable");
 
         let result = self
             .io
@@ -266,9 +244,9 @@ impl UdpSocket {
 
         match &result {
             Ok((size, addr, ts)) => {
-                trace!(size, ts = debug(ts), addr = debug(addr), "received message")
+                tracing::trace!(size, ts = ?ts, addr = ?addr, "received message");
             }
-            Err(e) => debug!(error = debug(e), "error receiving data"),
+            Err(e) => tracing::debug!(error = ?e, "error receiving data"),
         }
 
         result
@@ -303,13 +281,14 @@ fn recv(
 
             #[cfg(target_os = "linux")]
             ControlMessage::ReceiveError(_error) => {
-                warn!("unexpected error message on the MSG_ERRQUEUE");
+                tracing::warn!("unexpected error message on the MSG_ERRQUEUE");
             }
 
             ControlMessage::Other(msg) => {
-                warn!(
+                tracing::warn!(
                     "weird control message {:?} {:?}",
-                    msg.cmsg_level, msg.cmsg_type
+                    msg.cmsg_level,
+                    msg.cmsg_type
                 );
             }
         }
@@ -351,26 +330,29 @@ fn fetch_send_timestamp_help(
                 // the timestamping does not set a message; if there is a message, that means
                 // something else is wrong, and we want to know about it.
                 if error.ee_errno as libc::c_int != libc::ENOMSG {
-                    warn!(
+                    tracing::warn!(
                         expected_counter,
-                        error.ee_data, "error message on the MSG_ERRQUEUE"
+                        error.ee_data,
+                        "error message on the MSG_ERRQUEUE"
                     );
                 }
 
                 // Check that this message belongs to the send we are interested in
                 if error.ee_data != expected_counter {
-                    warn!(
+                    tracing::debug!(
                         error.ee_data,
-                        expected_counter, "Timestamp for unrelated packet"
+                        expected_counter,
+                        "Timestamp for unrelated packet"
                     );
                     return Ok(None);
                 }
             }
 
             ControlMessage::Other(msg) => {
-                warn!(
+                tracing::warn!(
                     msg.cmsg_level,
-                    msg.cmsg_type, "unexpected message on the MSG_ERRQUEUE",
+                    msg.cmsg_type,
+                    "unexpected message on the MSG_ERRQUEUE",
                 );
             }
         }
@@ -533,25 +515,30 @@ mod tests {
         let t2 = t2.unwrap();
         let delta = t2 - t1;
 
-        assert!(delta.to_seconds() > 0.15 && delta.to_seconds() < 0.25);
+        // this can be flaky on freebsd
+        assert!(
+            delta.to_seconds() > 0.15 && delta.to_seconds() < 0.25,
+            "delta was {}s",
+            delta.to_seconds()
+        );
     }
 
     #[tokio::test]
     #[cfg(target_os = "linux")]
     async fn timestamping_reasonable_so_timestamping() {
-        timestamping_reasonable(TimestampMethod::SoTimestamping, 8000, 8001).await
+        timestamping_reasonable(TimestampMethod::SoTimestamping, 8000, 8001).await;
     }
 
     #[tokio::test]
     #[cfg(target_os = "linux")]
     async fn timestamping_reasonable_so_timestampns() {
-        timestamping_reasonable(TimestampMethod::SoTimestampns, 8002, 8003).await
+        timestamping_reasonable(TimestampMethod::SoTimestampns, 8002, 8003).await;
     }
 
     #[tokio::test]
     #[cfg(unix)]
     async fn timestamping_reasonable_so_timestamp() {
-        timestamping_reasonable(TimestampMethod::SoTimestamp, 8004, 8005).await
+        timestamping_reasonable(TimestampMethod::SoTimestamp, 8004, 8005).await;
     }
 
     #[tokio::test]
